/***********************************************************************

  webtty.js - client side script used to display output from server and 
              to send keypresses to server

  This file is part of the webtty package. The webtty package was
  written by Martin Steen Nielsen

  Development is hosted by testape.com at www.testape.com/webtty

  webtty is free software; you can redistribute it and/or modify it 
  under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  webtty is distributed in the hope that it will be useful,but 
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  
  See the GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with webtty; if not, write to the Free Software Foundation, 
  Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

***********************************************************************/
var webTtyTaskArray= new Array;


function fullURIEncode(key)
{
  // return encodeURICOmponent(key);
  return "%"+("00" + key.toString(16)).match(/([0-9A-Fa-f][0-9A-Fa-f])$/)[0];
}


/***********************************************************************
   closes terminal
***********************************************************************/
function wclose(id)
{
  var elem = document.getElementById(id);
  elem.style.display='none';
  return false;
}




/***********************************************************************
   constructor webtty is called from client. Constructor expects to 
   receive id of element that will hold the html for the new process. 
   Function exit_func(err) is optional, but if given, it will be called 
   on errors or if the server terminates the process early.

   Sample usage -->

   <html>
     <script type="text/javascript" src="webtty.js">
     <body onLoad="new webtty('webtty_element');">
       <div id="webtty_element"></div>
     </body>
   </html>


***********************************************************************/

function zeroOrText(s){
  if (s == "\0")
            return "";
  else
    return s;
};

function newStrLen(l) {
    var s='';
    while (l>0) {
        s = s +' ';
        l--;
    }
    return s;
};

var capPaste = false;

function webtty(receiving_elementID, exit_func)
{
    this.getRange = function () {
        var ctrl = this.output_area;
        if(ctrl.setSelectionRange)
	      {
		        ctrl.focus();
		        ctrl.setSelectionRange(pos,pos);
	      }
	      else if (ctrl.createTextRange) {
		        var range = ctrl.createTextRange();
		        range.collapse(true);
		        range.moveEnd('character', pos);
		        range.moveStart('character', pos);
		        range.select();
	      }
    };
    this.setCursorTo = function (pos) {
        var ctrl = this.output_area;
        if(ctrl.setSelectionRange)
	      {
		        ctrl.focus();
		        ctrl.setSelectionRange(pos,pos);
	      }
	      else if (ctrl.createTextRange) {
		        var range = ctrl.createTextRange();
		        range.collapse(true);
		        range.moveEnd('character', pos);
		        range.moveStart('character', pos);
		        range.select();
	      }
    };
    this.setCursor = function (){
        var ctrl = this.output_area;
        if(ctrl && (ctrl.xfocused || ctrl.focused) && this.tty) {
            while(this.tty.buffer.length>0 && typeof(this.tty.buffer[this.tty.buffer.length-1])=='undefined') {
                this.tty.buffer.length=this.tty.buffer.length-1;
            }
            var buf =  this.tty.buffer;
            var x = this.tty.x;
            var y = this.tty.y + this.tty.offset;
            var pos = 0;
            for (var i =0; i<y-1; i++) {
                pos = pos + buf[i].length + ctrl.sepl;
            }
            pos = pos + x -1;
            this.setCursorTo(pos);
	      }
    }

    this.tty = {
        'node' : this,
        'buffer':[''],
        'x' : 1,
        'y' : 1,
        'fill':0,
        'offset' : 0,
        'max_x' : 80,
        'max_y' : 24,
        'scroll':{
            'top': 1,
            'bottom' : 24
        },
        'wrap' : true,
        'fn':{
            'ins' : function (tty, i, size) {
                for(var n=tty.buffer.length+size; n>=tty.buffer.i ;n--){
                    if(n>=tty.buffer.i + size)
                        tty.buffer[n] = tty.buffer[n-size];
                    else
                        tty.buffer[n] = '';
                };
            },

            'shift': function (tty, top, size, delta) {
                var tmp = [];
                for(var n = 0; n<size && size>0 ; n++){
                    tmp[n] = tty.buffer[top+n];
                    tty.buffer[top+n] = '';
                }
                for(var n = 0; n<size && size>0 ; n++){
                    tty.buffer[top+n + delta] = tmp[n];
                }
            },
            'del'  : function (tty, top, size){
                tty.buf.shift(tty,top + size, -size);
                tty.buffer.length = tty.buffer.length - size;
            },
            'clear' :function (tty, top, size) {
                for(var n = top; n<top+size ; n++){
                    tty.buffer[n] = '';
                }
            },
            'fixY' :function (tty, y) {
                for(var n = tty.fill; n<=y ; n++){
                    tty.buffer[n] =  (tty.buffer[n] ?  tty.buffer[n] : '');
                }
                tty.fill = y;
            },
            'fixX' :function (tty, y, x) {
                tty.fn.fixY(tty, y);
                while (x>tty.buffer[y].length) {
                    tty.buffer[y] =  tty.buffer[y] + ' ';
                }
            },
            'nil'  : function (tty, re_a) {
            },
            'clearLineFromPos': function (tty) {
                tty.buffer[tty.y + tty.offset - 1] = tty.buffer[tty.y + tty.offset - 1].substring(0,tty.x-1);
            } ,
            'clearLineBeforePos': function (tty) {
                tty.buffer[tty.y + tty.offset - 1] = newStrLen(tty.x) + tty.buffer[tty.y + tty.offset - 1].substring(tty.x);
            },
            'clearLine':function (tty) {
                tty.fn.clear(tty, tty.y + tty.offset, 1);
            },
            'clearScreenFromPos': function (tty) {
                tty.fn.clearLineFromPos(tty);
                tty.fn.clear(tty, tty.offset + tty.y, tty.max_y - tty.y );
            },
            'clearScreenBeforePos': function (tty) {
                tty.fn.clearLineBeforePos(tty);
                tty.fn.clear(tty, tty.offset, tty.y);
            },
            'clearScreen': function (tty) {
                tty.fn.clear(tty, tty.offset,tty.max_y)
            },

            'saveCursorPos':function (tty) {
                tty.save_pos = {'x' : tty.x, 'y' : tty.y};
            },
            'restoreCursorPos':function (tty) {
                tty.x = tty.save_pos.x;
                tty.y = tty.save_pos.y;
            },
            'setCursorPos' : function (tty, re_a) {
                var n = Number(re_a[1] ? re_a[1] : 1);
                while(tty.y <= n){
                    tty.buffer[tty.y + tty.offset - 1] = (tty.buffer[tty.y + tty.offset - 1] ? tty.buffer[tty.y + tty.offset - 1] : '');
                    tty.y++;
                }
                tty.y = n;
                tty.x = Number(re_a[3] ? re_a[3] : 1);
            },
            'moveCursorUp'  : function (tty, re_a) {
                tty.y = tty.y - Number(typeof(re_a[1]) != 'undefined' ? re_a[1] : 1);
                tty.y = Number(tty.y < 1 ? 1 : tty.y);
            }, 
            'moveCursorDown' :function (tty, re_a) {
                var n = Number(typeof(re_a[1]) != 'undefined' ? re_a[1] : 1);
                while (n>0) {
                    tty.y ++;
                    if ( tty.y >  tty.max_y ){
                        tty.y = tty.max_y;
                    }
                    n--
                    tty.buffer[tty.y + tty.offset - 1] = (tty.buffer[tty.y + tty.offset - 1] ? tty.buffer[tty.y + tty.offset - 1] : '');
                }
            },
            'moveCursorToRow'  : function (tty, re_a) {
                var n = Number(re_a[1] ? re_a[1] : 1);
                if (tty.y >= n) {
                    re_a[1] =  tty.y - n;
                    tty.fn.moveCursorUp(tty, re_a);
                } else {
                    re_a[1] =  n - tty.y;
                    tty.fn.moveCursorDown(tty, re_a);
                };
            },
            'moveCursorRight' :function (tty, re_a) {
                var n = Number(typeof(re_a[1]) != 'undefined' ? re_a[1] : 1);
                          tty.x = tty.x + n;
                          tty.x = Number(tty.x > tty.max_x ? tty.max_x :tty.x); //!!!check tty!!!
            },
            'moveCursorToColumn' :function (tty, re_a) {
                          var n = Number(re_a[1] ? re_a[1] : 1);
                          tty.x = n;
                          tty.x = Number(tty.x > tty.max_x ? tty.max_x :tty.x); //!!!check tty!!!
            },
            'moveCursorLeft' :function (tty, re_a) {
                          var n = Number(typeof(re_a[1]) != 'undefined' ? re_a[1] : 1);
                          tty.x = tty.x - n;
                          tty.x = Number(tty.x < 1 ? 1 :tty.x);
            },
            'insertLine' :function (tty, re_a) {
                var n = Number(re_a[1] ? re_a[1] : 1);
                if(tty.scroll.bottom - tty.y - n > 0) {
                    tty.fn.shift(tty,tty.offset + tty.y - 1,tty.scroll.bottom - tty.y - n + 1, n);
                } else {
                    tty.fn.clear(tty,tty.offset + tty.y - 1,tty.scroll.bottom - tty.y);
                }
            },
            'deleteLine' :function (tty, re_a) {
                var n = Number(re_a[1] ? re_a[1] : 1);
                tty.fn.shift(tty,tty.offset + tty.y + n - 1,tty.scroll.bottom - tty.y, -n);
            },
            'deleteChar' :function (tty, re_a) {
                var n = Number(re_a[1] ? re_a[1] : 1);
                while (n>0) {
                    tty.buffer[tty.y + tty.offset - 1] = tty.buffer[tty.y + tty.offset - 1].substring(0, tty.x - 1) + tty.buffer[tty.y + tty.offset - 1].substring(tty.x);
                    n--;
                }
            },
            'eraseChar' :function (tty, re_a) {
                var n = Number(re_a[1] ? re_a[1] : 1);
                while (n>0) {
                    tty.buffer[tty.y + tty.offset - 1] = tty.buffer[tty.y + tty.offset - 1].substring(0, tty.x - 1) + " " + tty.buffer[tty.y + tty.offset - 1].substring(tty.x);
                    n--;
                }
            },
            'setScrollRegion' :function (tty, re_a) {
                var top = Number(re_a[2] ? re_a[2] : 1);
                var bottom = Number(re_a[4] ? re_a[4] : tty.max_y);
                tty.scroll.top = top;
                tty.scroll.bottom = bottom;
            },
            'scrollRegionDown' :function (tty) {
                tty.fn.shift(tty,
                             tty.scroll.top + tty.offset + 1,
                             tty.scroll.bottom - tty.scroll.top - 1,
                             -1);
            },
            'scrollRegionUp' :function (tty) {
                tty.fn.shift(tty,
                             tty.scroll.top + tty.offset,
                             tty.scroll.bottom - tty.scroll.top - 1,
                             1);
            },
            'resetTerminal' : function (tty) {/*  TODO reset terminal*/;
                tty.x = 1;
                tty.y = 1;
                tty.max_x = 80;
                tty.max_y = 24;
                tty.scroll.top = 1;
                tty.scroll.bottom = 24;
                tty.wrap = true;
                tty.fn.clear(tty, tty.buffer.length, tty.max_y);
                tty.offset = tty.buffer.length - tty.max_y;
            }
        }
    };
    this.fastRestart = false;
    this.ctrlKey = false
    this.receiving_elementID = receiving_elementID;
    /***********************************************************************
       Creates XMLHTTPRequest object
    ***********************************************************************/
    this.create_transfer_object = function()
    {
        var Request = false;

        if (window.XMLHttpRequest)
        {
	          /*Gecko- Safari, Konqueror*/
	          Request = new XMLHttpRequest();
        }
        else if (window.ActiveXObject)
        {
	          /*Internet explorer*/
	          try
            {
                Request = new ActiveXObject("Microsoft.XMLHTTP");
            }
	          catch (CatchException)
            {
                Request = new ActiveXObject("Msxml2.XMLHTTP");
            }
        };

        if (!Request)
        {
	          alert("Can`t create XMLHttpRequest");
        };

        return Request;
    }
  
    /*******************************************************
     REGEXP FUNCTIONS
    *******************************************************/
    this.escRegExp = [{
        'r':/^(\x1b[\x00-\x1f]*)/,
        'n':[
            {
                'r':/^([\[])/,
                'n':[
                    
                    { 'r':/^(([?]?[0-9][0-9]?)[hl]|O|N)/,
                      // Application/Numeric  keypad mode
                      'f': this.tty.fn.nil
                    },
                    { 'r':/^((;)?([0-9][0-9]?)?)*m/,
                      // set color params
                      'f': this.tty.fn.nil
                    },
                    { 'r':/^(c)/,//Requests a Report Device Code response from the device.
                      'f':function (tty, re_a) { tty.node.buffer_next_data("\x1b[00c");}},
                    { 'r':/^(5n)/,//Requests a Report Device Status response from the device.
                      'f':function (tty, re_a) { tty.node.buffer_next_data("\x1b[0n");}},
                    { 'r':/^(6n)/,//Requests a Report Cursor Position response from the device.
                      'f':function (tty, re_a) { tty.node.buffer_next_data("\x1b["+(tty.y)+";"+(tty.x)+"R");}},
                    { 'r':/^(7h)/,//Line WRAP ON
                      'f':function (tty) { tty.wrap = true;}},
                    { 'r':/^(7l)/,//Line WRAP OFF
                      'f':function (tty) { tty.wrap = false;}},
                    { 'r':/^([1-9][0-9]?)?(;([1-9][0-9]?[0-9]?)?)?[Hf]/,
                      // Set cursor position
                      'f': this.tty.fn.setCursorPos
                    },
                    { 'r':/^([1-9][0-9]?)?A/,
                      // Move cursor UP
                      'f': this.tty.fn.moveCursorUp
                    },
                    { 'r':/^([1-9][0-9]?)?B/,
                      // Move cursor DOWN
                      'f': this.tty.fn.moveCursorDown
                    },
                    { 'r':/^([1-9][0-9]?)?C/,
                      // Move cursor right
                      'f': this.tty.fn.moveCursorRight
                    },
                    { 'r':/^([1-9][0-9]?)?d/,
                      // Move cursor to row
                      'f': this.tty.fn.moveCursorToRow
                    },
                    { 'r':/^([1-9][0-9]?)?D/,
                      // Move cursor left
                      'f': this.tty.fn.moveCursorLeft
                    },
                    { 'r':/^([1-9][0-9]?)?G/,
                      // Move cursor to column
                      'f': this.tty.fn.moveCursorToColumn
                    },
                    { 'r':/^([1-9][0-9]?)?L/,
                      // Insert Line
                      'f': this.tty.fn.insertLine
                    },
                    { 'r':/^([1-9][0-9]?)?M/,
                      // Delete Line
                      'f':this.tty.fn.deleteLine
                    },
                    { 'r':/^([1-9][0-9]?)?[P]/,
                      // Delete Char
                      'f':this.tty.fn.deleteChar
                    },
                    { 'r':/^s/, //save cursor pos
                      'f':this.tty.fn.saveCursorPos
                    },
                    { 'r':/^u/, //restore cursor pos
                      'f':this.tty.fn.restoreCursorPos
                    },
                    { 'r':/^(([1-9][0-9]?)?(;([1-9][0-9]?[0-9]?))?r)/,
                      // Set scroll region
                      'f': this.tty.fn.setScrollRegion
                    },
                    { 'r':/^[0]?g/,
                      'f': function (tty) { tty.tab[tty.x] = null; }
                    },
                    { 'r':/^3g/,
                      'f': function (tty) { tty.tab = []; }
                    },
                    { 'r':/^[0]?K/,
                      'f': this.tty.fn.clearLineFromPos
                    },
                    { 'r':/^1K/,
                      'f': this.tty.fn.clearLineBeforePos
                    },
                    { 'r':/^2K/,
                      'f': this.tty.fn.clearLine
                    },
                    { 'r':/^[0]?J/,
                      'f':  this.tty.fn.clearScreenFromPos
                    },
                    { 'r':/^1J/,
                      'f':  this.tty.fn.clearScreenBeforePos
                    },
                    { 'r':/^2J/,
                      'f': this.tty.fn.clearScreen
                    },
                    { 'r':/^([1-9][0-9]?)?X/,
                      'f': this.tty.fn.eraseChar
                    },
                    {
                        'r':/^[?][^c]*c/,
                        'f':this.tty.fn.nil
                    }/*,
                    {
                        'r':/^(.*)/, // Unknown escape
                        'f':function (tty, re_a) { alert(re_a[1]); }
                    }*/
                ]
            },
            { 'r':/^([=><])/,
              'f':this.tty.fn.nil
            },
            { 'r':/^(c)/,/*  TODO reset terminal*/
              'f': this.tty.fn.resetTerminal
            },
            { 'r':/^([\(])[\x40-\x7e]/, /*  TODO set default font*/
              'f':this.tty.fn.nil
            },
            { 'r':/^([\)])[\x40-\x7e]/, /*  TODO set alternate font*/
              'f':this.tty.fn.nil
            },
            { 'r':/^([\)\(])[BO0UK]/, /*  TODO Start sequence defining G0/G1 character set*/
              'f':this.tty.fn.nil
            },
            { 'r':/^7/, //save cursor pos
              'f': this.tty.fn.saveCursorPos
            },
            { 'r':/^8/, //restore cursor pos
              'f': this.tty.fn.restoreCursorPos
            },
            { 'r':/^D/, // Scroll down
              'f': this.tty.fn.scrollRegionDown
            },
            { 'r':/^M/, // Scroll up
              'f': this.tty.fn.scrollRegionUp
            },
            { 'r':/^H/, // Set TAB position
              'f':function (tty) { tty.tab[tty.x] = true;}
            },
            { 'r':/^Z/, // ident
//              'f':function (tty) {  tty.node.buffer_next_data("\x1b\\Z");}
              'f':function (tty) {  tty.node.buffer_next_data("\x1b[?1c");}
            }/*,
            {
                'r':/^(.*)/, // Unknown escape
                'f':function (tty, re_a) { alert(re_a[1]); }
            }*/
        ]
    }];

    this.runRegexp = function(data, escRegExp_a) {
        for (var i=0; i<escRegExp_a.length; i++ ) {
            if (data.search(escRegExp_a[i].r) == 0) {
                var regRes = escRegExp_a[i].r.exec(data);
                if (escRegExp_a[i].f) {
                    escRegExp_a[i].f(this.tty, regRes);
                    return regRes[0].length;
                } else if (escRegExp_a[i].n) {
                    var fRes = this.runRegexp(data.substring(regRes[0].length), escRegExp_a[i].n);
                    return  ( fRes >0 ? regRes[0].length + fRes : 0);
                };
            }
        }
        return 0;
    }

    this.makeEscape = function (data) {
        var tty = this.tty;
        while (data.length > 0) {
            var c = data.charAt(0);
            try {
                var esc = (c=='\x1b' ?  this.runRegexp(data, this.escRegExp) : 0);
                if (esc>0) {
                    data = data.substring(esc);
                } else if(c=='\n' /* 0x0A  */ || c=='\x0B' || c=='\x0C') {
                    data = data.substring(1);
                    tty.y++;
                    if (tty.y > tty.scroll.bottom) {
                        tty.y --;
                        if  (tty.scroll.top == 1) {
                            //scroll buffer
                            tty.fn.shift(tty,
                                         tty.offset + tty.scroll.bottom,
                                         tty.max_y - tty.scroll.bottom,
                                         1);
                            tty.offset++;
                        }
                        else {
                            //scroll region
                            tty.fn.shift(tty,
                                         tty.offset + tty.scroll.top,
                                         tty.scroll.bottom - tty.scroll.top,
                                         -1);
                        }
                    } else {
                        if (tty.y>tty.max_y) {
                            tty.fn.shift(tty,
                                         tty.offset + tty.y,
                                         tty.scroll.bottom - tty.y -1,
                                         1);
                        }
                    }
                    tty.fn.fixX(tty, tty.offset + tty.y - 1, tty.x - 1);
                }else if(c=='\x08') {
                    data = data.substring(1);
                    if(tty.x>1) {
                        tty.x --;
                    }
                }else if(c=='\r' /* 0x0D */) {
                    data = data.substring(1);
                    tty.x = 1;
                }else if(c>='\x00' && c<='\x1F' ) {
                    data = data.substring(1);
                } else {
                    data = data.substring(1);
                    tty.fn.fixX(tty, tty.offset + tty.y - 1, tty.x - 1);
                    tty.buffer[tty.y + tty.offset - 1] = tty.buffer[tty.y + tty.offset - 1].substring(0, tty.x - 1) + c + tty.buffer[tty.y + tty.offset - 1].substring(tty.x - 1 + 1);
                    tty.x = tty.x+1;
                    if(tty.x > tty.max_x && ! tty.wrap) {
                        tty.x = tty.max_x;
                    }
                };

                   if(0) {
                //!!!!!! DEBUG ONLY
                this.output_area.value = tty.buffer.join('\n');
                this.output_area.scrollTop=this.output_area.scrollHeight;
                }
            } catch(err) {
                data = data.substring(1);
                tty.fn.fixX(tty, tty.offset + tty.y - 1, tty.x - 1);
                tty.buffer[tty.y + tty.offset - 1] = tty.buffer[tty.y + tty.offset - 1].substring(0, tty.x - 1) + c + tty.buffer[tty.y + tty.offset - 1].substring(tty.x - 1 + 1);
                tty.x = tty.x+1;
                if(tty.x > tty.max_x && ! tty.wrap) {
                    tty.x = tty.max_x;
                }
            }
        };
        return tty.buffer.join('\n');
    };


    /***********************************************************************
       callback function whenever html data is ready
    ***********************************************************************/
    this.get_html_state = function()
    {
        if (this.xhr_out.readyState == 4)
        {
            //alert("status:"+this.xhr_out.status);
            if ((this.xhr_out.status!=200)
                &&(this.xhr_out.status!=204)
                &&(this.xhr_out.status!=0)
                &&(typeof(this.xhr_out.status) != 'undefined'))
                this.error();
            if (this.xhr_out.status==200 || this.xhr_out.status==204 || (typeof(this.xhr_out.status) == 'undefined')) {
                this.initCounter = this.initCounter + 1;
                //alert('responseText:'+zeroOrText(this.xhr_out.responseText));
                hiddenDiv.innerHTML = zeroOrText(xmlTransform(this.xhr_out));
                //alert('after\10'+hiddenDiv.innerHTML)
                execScripts(hiddenDiv);
                var response = htmlDecode(getText(hiddenDiv));
                if(this.initCounter==1) {
                    this.element.innerHTML = "<img title=\"header=[webUpdating] body=[]\" class=\"ttyactivityon\" src=\"/loading.gif\" alt=\"\">" + "<textarea id=webtty_output rows=24 cols=80 xondblclick=\"UnHide(this);\""+" onblur=\"this.focused=false;\"  onfocus=\"this.focused=true;\" onclick=\"this.me.setCursor();\" onmousedown=\"this.me.setCursor();\" onmouseup=\"this.me.setCursor();\">"+"</textarea>";
                    this.session_id  = ":)";
                    this.activity_img = findChildByTag(this.element, 'img', 3, "", "");
                    this.output_area = findChildByTag(this.element, 'textarea', 3, "", "");
                    this.output_area.id = 'default';
                    this.output_area.me = this;
                    this.output_area.value = " \n \n \n \n \n ";
                                           //" \nr \nr \nr \nr \nr "
                    var ret;
                    if( this.output_area.createTextRange) {
                        var rng = this.output_area.createTextRange();
                        rng.collapse(true);
		                    rng.moveEnd('character', 5);
		                    rng.moveStart('character', 0);
                        //      
                        ret = rng.text;
                    } else {
                        ret = this.output_area.value.substr(0,5);
                    };
                    ret = ret.replace(/[^ ]/gi,"");
                    ret = ret.length;
                    this.output_area.sepl = (ret>2 ? 1 :2);// (this.output_area.value.length-4)/4;
                    if (response > '') this.handle_data(response);
                    //              alert(this.output_area.value);
                    var obj = this;
                    this.output_area.onkeypress = function(e) { if (obj.running) return obj.key_event(e); };
                    this.output_area.onkeydown  = function(e) { if (obj.running) return obj.special_key_event(e); };
                    $(this.output_area).bind('cut', function(e) {
                        return false;
                    });
                    $(this.output_area).bind('paste', function(e) {
                        return false;
                    });
                    $(this.output_area).bind('input', function(e) {
                        return false;
                    });

                    this.output_area.xfocused = false;
                    this.output_area.hasFocus = function() {
                        return obj.output_area.xfocused;
                    };
                    $(this.output_area).bind('focus', function() {
                        obj.output_area.xfocused=true;
                    });
                    $(this.output_area).bind('blur', function() {
                        obj.output_area.xfocused=false;
                    });

                    this.activity_img.className = "ttyactivityoff";
                    this.timerID=setTimeout(function() {
                        if(webTtyTaskArray[obj.receiving_elementID] != null)
                            webTtyTaskArray[obj.receiving_elementID].get_data();
                    } , 200);
                }
                else {
                    if(response!='') handle_data(response);
                }
            }
        }
    }

    /***********************************************************************
       Request a new process and HTML data from server
    ***********************************************************************/
    this.get_html = function()
    {
        var obj = this;
        var path = encodeURIComponent(buildPath(findParentByTag(this.element, "LI", 10), ","));
        //      if (null == this.xhr_out)
        this.xhr_out = SendRequestAsync('POST',
                                        PAGENAME,
                                        'command=tty&path={'+path+'}&value=&',
                                        function() { if (obj.running) obj.get_html_state(); },
                                        "retry"
                              );
    }




    /***********************************************************************
       callback function whenever process data is ready
    ***********************************************************************/
    this.get_data_state = function()
    {
        if (this.xhr_out.readyState == 4)
        {
            this.activity_img.className = "ttyactivityoff";
            if ((this.xhr_out.status!=200)&&(this.xhr_out.status!=204)&&(this.xhr_out.status!=0)
                &&(typeof(this.xhr_out.status) != 'undefined'))
                this.error();
            //alert( "rcv:"+zeroOrText(this.xhr_out.responseText));
            hiddenDiv.innerHTML = zeroOrText(xmlTransform(this.xhr_out));
            //alert("html:"+hiddenDiv.innerHTML );
            execScripts(hiddenDiv);
            var data = htmlDecode(getText(hiddenDiv));
            if (data == '_C_L_O_S_E_D_')
            {
                //                this.output_area.value = this.output_area.value +  data;
           /*     this.output_area.value = this.makeEscape(data);
                this.output_area.scrollTop=this.output_area.scrollHeight;
                this.setCursor();*/
                this.handle_data(data);
                var rows = this.output_area.value.split("\n").length;
                this.output_area.rows = (rows > 5 ? (rows < 24 ? rows : 24) : 5);
                this.session_id=null;
                if (undefined != this.webtty_exit)
                    this.webtty_exit(0);
                this.cleanup();
            }
            else
            {
                if (data > '')
                    this.handle_data(data);
                var obj =this;
                this.timerID=setTimeout(function() {
                    obj.activity_img.className = "ttyactivityoff";
                    if(webTtyTaskArray[obj.receiving_elementID] != null)
                        webTtyTaskArray[obj.receiving_elementID].get_data();
                }, this.fastRestart ? 10 : this.timerDelay);
                this.fastRestart = false;
                this.timerDelay = 1000;
                //this.get_data();
            }
        }
    }


    /***********************************************************************
       Request new data from the server
    ***********************************************************************/
    this.get_data = function()
    {
        if (this.timerID != null) {
            clearTimeout( this.timerID);
            this.activity_img.className = "ttyactivityoff";
        }
        this.timerID = null;
        var obj = this;
        var out;
        var k;
        var path = encodeURIComponent(buildPath(findParentByTag(this.element, "LI", 10), ","));
        k = this.kbd_buffer.search(/[%][%][%]/);
        if(k == 0) {
            out = "";
            this.kbd_buffer = this.kbd_buffer.substr(3);
        }else if(k>0) {
            out = "&value="+this.kbd_buffer.substr(0,k);
            this.kbd_buffer = this.kbd_buffer.substr(k);
        } else {
            out = "&value="+this.kbd_buffer;
        };
        this.activity_img.className = "ttyactivityon";
        this.xhr_out = SendRequestAsync('POST',
                                       PAGENAME,
                                       'command=tty&path={'+path+'}&'+ out + '&',
                                       function() { return obj.get_data_state(); },
                                        "retry"
                                      );
        this.kbd_buffer = '';
    }


    /***********************************************************************
       exit page. Remove handles and terminate session
    ***********************************************************************/
    this.cleanup = function()
    {
        this.running     = false;
        if (this.timerID) {
            clearTimeout( this.timerID);
            this.activity_img.className = "ttyactivityoff";
        }
        this.timerID = null;
        webTtyTaskArray[this.receiving_elementID] = null;
        if (this.xhr_kbd)
        {
            this.xhr_kbd.onreadystatechange = function() {};
            this.xhr_kbd.abort();
            delete this.xhr_kbd;
        }
        if (this.xhr_out)
        {
            this.xhr_out.onreadystatechange = function() {};
            this.xhr_out.abort();
            delete xhr_out;
        }
        if (null != this.session_id)
        {
            var path = encodeURIComponent(buildPath(findParentByTag(this.element, "LI", 10), ","));
            var exit_req = SendRequestAsync('POST',
                                            PAGENAME,
                                            'command=tty&path={'+path+'}&',
                                            function() {}
                                           );
            delete exit_req;
            this.session_id=null;
        }
        if (null != this.tty)
            this.tty = null;
    }
    
    /***********************************************************************
       handles process data
    ***********************************************************************/
    this.handle_data = function(data)
    {
        this.output_area.value = this.makeEscape(data);
        this.output_area.scrollTop=this.output_area.scrollHeight;
        this.setCursor();
    }

    /***********************************************************************
      buffer data and send to server
    ***********************************************************************/
    this.buffer_next_data = function(cmd)
    {
        this.kbd_buffer += cmd;
        if (this.timerID != null) {
            clearTimeout( this.timerID);
            this.activity_img.className = "ttyactivityoff";
            var obj = this;
            this.fastRestart = true;
        this.timerID = setTimeout(function() {
             webTtyTaskArray[obj.receiving_elementID].get_data();
          }, 10);
      } else {
        this.timerDelay = 10;
      }
    }



  /***********************************************************************
       handles normal keypresses
  ***********************************************************************/
  this.key_event = function(e)
    {
        if (!e) var e = window.event;
        if (e.keyCode) key = e.keyCode;
        else if (e.which) key = e.which;
        if ( e != window.event && typeof(e.charCode)=="undefined")
            return true;
        /*handle escape symbols for Ctrl X*/
        if ((String.fromCharCode(key) == "^") && !this.ctrlKey) {
            this.ctrlKey = true;
            return false;
        }
        if (this.ctrlKey) {
            this.ctrlKey = false;
            if ( (key > 64) && (key < 96) && (key != 94) )
                key = key - 64;
        }
        
        if (key==13 || key == 10 || key == 9 || key == 8) {
            return false;
        }

        this.buffer_next_data(fullURIEncode(key));
        return false;
    }


  /***********************************************************************
       handles special keypresses
  ***********************************************************************/
  this.special_key_event = function (e)
    {
      if (!e) var e = window.event;
      if (e.keyCode) key = e.keyCode;
      else if (e.which) key = e.which;

     
      if (key==9)
        { 
          this.buffer_next_data(escape("\t")); 
          return false; 
        }
      else if (key==8)
        { 
          this.buffer_next_data(escape("\010")); 
          return false; 
        }
      else if (key == 40)
        { 
          this.buffer_next_data(escape("\033[B")); 
          return false; 
        }
      else if (key == 38)
        { 
          this.buffer_next_data(escape("\033[A")); 
          return false; 
        }
      else if (key == 38)
        { 
          this.buffer_next_data(escape("\033[A")); 
          return false; 
        }
      else if (key == 13 || key == 10)
        {
          this.buffer_next_data(escape("\r"));
          this.cr = true;
          return false; 
        }
    }


  /***********************************************************************
       handles errors
  ***********************************************************************/
  this.error = function()
    {
      if (undefined != this.webtty_exit)
        this.webtty_exit(1);
      if (this.element)
        this.element.innerHTML="Error";
      this.cleanup();
    }


  this.element =  document.getElementById(receiving_elementID);
  if (null == this.element)
    return this.error();

  this.running     = true;
  this.session_id  = null;
  this.output_area = null;
  this.kbd_buffer  = '';
  this.webtty_exit = exit_func;
  this.timerID = null;
  this.timerDelay = 1000;
  this.initCounter = 0;
  webTtyTaskArray[receiving_elementID] = this;
}


function init_webtty_object(objID, exit_handler) {
  /*if (undefined != exitFunc) 
    exit_handler = exitFunc;
    else
    exit_handler =  function webtty_exit()
    {
    var div_element = obj;
    div_element.innerHTML = 'Refresh page to get a new virtual machine';
    };*/
  a = new webtty(objID, exit_handler);
  a.get_html();
  return a;
}
